/**
*
* A promela grammar for ANTLR 4 derived from Spin documentation
* available at http://spinroot.com/spin/Man/grammar.html
*
*/
grammar Promela;

spec	: module+;

module	: proctype	 /* proctype declaration */
	    | init		 /* init process       */
	    | never		 /* never claim        */
	    | trace		 /* event trace        */
	    | utype		 /* user defined types */
	    | mtype		 /* mtype declaration  */
	    | declList /* global vars, chans */
	    ;

proctype: active? PROCTYPE name LPAR declList? RPAR priority? enabler? LBRAKET sequence RBRAKET;

init	: INIT priority? LBRAKET sequence RBRAKET;

never	: NEVER	LBRAKET sequence RBRAKET;

trace	: TRACE LBRAKET sequence RBRAKET;

utype	: TYPEDEF name LBRAKET declList RBRAKET;

mtype	: MTYPE '='? LBRAKET name (',' name)* RBRAKET;

declList: oneDecl (';' oneDecl)*;

oneDecl: visible? typename  ivar (',' ivar)*;

typename: BIT | BOOL | BYTE | SHORT | INT | MTYPE | CHAN | uname;

active  : ACTIVE ('[' constant ']')?;	/* instantiation */

priority: PRIORITY constant;	/* simulation priority */

enabler : PROVIDED LPAR expr RPAR;	/* execution constraint */

visible	: HIDDEN_ | SHOW;

sequence: step ((SEMICOL|RARROW)? step)*;

step    : stmnt	(UNLESS stmnt)?
        | declList
	    | XR varref (',' varref)*
	    | XS varref (',' varref)*
	    ;

ivar    : name ('[' constant ']')? ('=' any_expr | '=' ch_init)?;

ch_init : '[' constant ']' OF '{' typename (',' typename)* '}';

varref	: name ('[' any_expr ']')? ('.' varref)?
        | PID;

send    : varref '!' send_args		/* normal fifo send */
	    | varref '!!' send_args	/* sorted send */
	    ;

receive : varref '?' recv_args		/* normal receive */
        | varref '??' recv_args	/* random receive */
	    | varref '?' '<' recv_args '>'	/* poll with side-effect */
	    | varref '??' '<' recv_args '>'	/* ditto */
	    ;

poll    : varref '?' '[' recv_args ']'	/* poll without side-effect */
	    | varref '??' '[' recv_args ']'	/* ditto */
	    ;

send_args: arg_lst | any_expr LPAR arg_lst RPAR;

arg_lst  : any_expr (',' any_expr)*;

recv_args: recv_arg (',' recv_arg)*  |  recv_arg LPAR recv_args RPAR;

recv_arg : varref | EVAL LPAR varref RPAR | '-'? constant;

assign  : varref EQ any_expr	/* standard assignment */
	    | varref INC	/* increment */
	    | varref DEC	/* decrement */
	    ;


stmnt	: IF stmnt_options FI		/* selection */                                 #IfStmnt
	    | DO stmnt_options OD		/* iteration */                                 #DoStmnt
	    | FOR LPAR range RPAR LBRAKET sequence RBRAKET		/* iteration */         #ForStmnt
	    | ATOMIC LBRAKET sequence RBRAKET	/* atomic sequence */                   #AtomicStmnt
	    | D_STEP LBRAKET sequence RBRAKET	/* deterministic atomic */              #DStepStmnt
	    | SELECT LPAR range RPAR	/* non-deterministic value selection */         #SelectStmnt
	    | LBRAKET sequence RBRAKET	/* normal sequence */                           #BraketStmnt
	    | send                                                                      #SendStmnt
	    | receive                                                                   #ReceiveStmnt
	    | assign                                                                    #AssignStmnt
	    | ELSE			/* used inside options */                                   #KeywordStmnt
	    | BREAK			/* used inside iterations */                                #KeywordStmnt
	    | GOTO name                                                                 #KeywordStmnt
	    | name ':' stmnt	/* labeled statement */                                 #LabelStmnt
	    | PRINT LPAR string (',' arg_lst)? RPAR                                     #PrintStmnt
	    | ASSERT expr                                                               #AssertStmnt
	    | expr			/* condition */                                             #ExprStmnt
	    ;

range	: varref ':' expr '..' expr
	    | varref IN varref
	    ;

stmnt_options : (DCOLON sequence)+;


andorOp	: '&&' | '||';

binaryOp	: '+' | '-' | '*' | '/' | '%' | '&' | '^' | '|'
	    | '>' | '<' | '>' '=' | '<=' | '==' | '!='
	    | '<<' | '>>' | andorOp;

unaryOp  : '~' | '-' | '!';

any_expr: LPAR any_expr RPAR                                                            #BraketExpr
        | any_expr binaryOp any_expr                                                    #BinaryExpr
	    | unaryOp any_expr                                                              #UnaryExpr
	    | LPAR any_expr RARROW any_expr ':' any_expr RPAR                               #AnyExpr
	    | LEN LPAR varref RPAR /* nr of messages in chan */                             #AnyExpr
	    | poll                                                                          #AnyExpr
	    | varref                                                                        #VarRefExpr
	    | constant                                                                      #ConstantExpr
	    | SKIP_                                                                         #Skip
	    | TIMEOUT                                                                       #Timeout
	    | NP_			/* non-progress system state */                                 #AnyExpr
	    | ENABLED LPAR any_expr RPAR  /* refers to a pid */                             #AnyExpr
	    | PC_VALUE LPAR any_expr RPAR /* refers to a pid */                             #AnyExpr
	    | name '[' any_expr ']' '@' name	/* refers to a pid */                       #AnyExpr
	    | RUN name LPAR arg_lst? RPAR priority?                                         #RunProctype
	    | GET_PRIORITY LPAR expr RPAR /* expr refers to a pid */                        #AnyExpr
	    | SET_PRIORITY LPAR expr ',' expr RPAR		/* first expr refers to a pid */    #AnyExpr
        ;

expr	: any_expr
    	| LPAR expr RPAR
    	| expr andorOp expr
	    | chanpoll LPAR varref RPAR	/* may not be negated */
        ;

chanpoll: FULL | EMPTY | NFULL | NEMPTY;

uname   : name;
name	: NameLiteral;

constant	: TRUE | FALSE | NumberLiteral;

string	: StringLiteral;

// LEXER

// Keywords

LPAR: '(';
RPAR: ')';
LBRAKET: '{';
RBRAKET: '}';
RARROW: '->';
SEMICOL: ';';
EQ: '=';
INC: '++';
DEC: '--';
DCOLON: '::';
PROCTYPE: 'proctype';
NEVER: 'never';
TRACE: 'trace';
TYPEDEF: 'typedef';
MTYPE: 'mtype';
BIT: 'bit';
BOOL: 'bool';
BYTE: 'byte';
SHORT: 'short';
INT: 'int';
CHAN: 'chan';
ACTIVE: 'active';
PRIORITY: 'priority';
PROVIDED: 'provided';
HIDDEN_: 'hidden';
SHOW: 'show';
UNLESS: 'unless';
XR: 'xr';
XS: 'xs';
OF: 'of';
EVAL: 'eval';
IF: 'if';
FI: 'fi';
DO: 'do';
OD: 'od';
FOR: 'for';
ATOMIC: 'atomic';
D_STEP: 'd_step';
SELECT: 'select';
ELSE: 'else';
BREAK: 'break';
GOTO: 'goto';
PRINT: 'print';
IN: 'in';
ASSERT: 'assert';
LEN: 'len';
TIMEOUT: 'timeout';
NP_: 'np_';
ENABLED: 'enabled';
PC_VALUE: 'pc_value';
RUN: 'run';
GET_PRIORITY: 'get_priority';
SET_PRIORITY: 'set_priority';
INIT: 'init';
TRUE: 'true';
FALSE: 'false';
SKIP_: 'skip';
FULL: 'full';
EMPTY: 'empty';
NFULL: 'nfull';
NEMPTY: 'nempty';
PID: '_pid';

// Literals

NameLiteral  : Alpha (Alpha | Digit)*;
StringLiteral: '"' StringCharacters? '"';
NumberLiteral: '0'
             | NonZeroDigit Digits?
             ;

WS: [ \t\r\n]+ -> skip;

COMMENT: '/*' .*? '*/' -> skip;
LINE_COMMENT: '//' ~[\r\n]* -> skip;

// Fragments

fragment StringCharacters:	StringCharacter+;
fragment StringCharacter:	~["\\\r\n];

fragment Alpha	: [a-zA-Z];
fragment NonZeroDigit: [1-9];
fragment Digits: Digit+;
fragment Digit	: [0-9];