package it.sagelab.dynaagen.core;

import it.sagelab.dynaagen.model.Proctype;
import it.sagelab.dynaagen.model.PromelaModel;
import it.sagelab.dynaagen.model.expr.Variable;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class DynaaGenerator {

    public enum Indentation {
        INC,
        DEC
    }

    private final PromelaModel model;
    private String outputDir = "gen/";
    private String filesPackage = "it.sagelab.dynaagen.autogen";
    private String globalVarsContainerClassName = "DataContainer";
    private int _indentation = 0;
    private PrintWriter printWriter;

    public DynaaGenerator(PromelaModel model) {
        this.model = model;
    }

    public void generate() throws IOException {
        generateDataContainer();
        for(Proctype proctype: model.getProctypes())
            generate(proctype);
    }

    public void setOutputDir(String outputDir) {
        this.outputDir = outputDir;
    }

    private void generate(Proctype proctype) throws IOException {
        String className = StringUtils.capitalize(proctype.getName()) + "Task";
        printWriter = new PrintWriter(outputDir + className + ".java");
        print( "package " + filesPackage + ";\n");
        print("import nl.tno.dynaa.model.functional.Task;", "import nl.tno.dynaa.model.functional.behavioural.*;");
        print("import nl.tno.dynaa.core.Event.EventType;", "");
        print(DynaaConsts.HEADER);
        print("public class " + className + " {");

        DynaaSegmentGenerator segmentGenerator = new DynaaSegmentGenerator(model);
        Set<DynaaSegment> segments = segmentGenerator.buildSegments(proctype);
        Set<String> events = new HashSet<>();
        for(DynaaSegment seg: segments) {
            events.addAll(seg.getNextSegments().keySet());
        }
        events.remove(DynaaSegment.SEGMENT_SUCCESS);
        print(Indentation.INC, "");
        for(String e: events) {
            print("public static final EventType " + e + " = Event.createType(\"" + e + "\", \"\");");
        }

        print("", "public static Task build(int _pid, " + globalVarsContainerClassName +" globalVars) {\n");
        print(Indentation.INC, "final TransitionBehaviourSpecification behaviour = new TransitionBehaviourSpecification();");
        print("final Task task = new Task(behaviour);");
        print("task.setProperty(\"GLOBAL_VARS\", globalVars);");
        for(Variable var: proctype.getParams().values()) {
            print("task.setProperty(\"" + var.getName() + "\", " + ExprToString.parse(var.getExpr()) +");");
        }

        print("");
        for(DynaaSegment seg: segments) {
            print("final TaskSegment " +seg.getName() + " = " + className + ".create" + StringUtils.capitalize(seg.getName()) + "();");
        }

        print("");
        print("behaviour.setInitialSegment(" + segmentGenerator.getSegment(proctype.getSequence()).getName() + ");");
        for(DynaaSegment seg: segments) {
            for(Map.Entry<String, DynaaSegment> e : seg.getNextSegments().entrySet()) {
                if(e.getValue() != null) {
                    print("behaviour.addTransition(" + seg.getName() + ", " + e.getKey() + ", " + e.getValue().getName() + ");");
                } else {
                    print("behaviour.addTransition(" + seg.getName() + ", " + e.getKey() + ", null);");
                }
            }
        }

        print("", "return task;");
        print(Indentation.DEC, "}\n");

        for(DynaaSegment seg: segments) {
            generateSegment(seg);
        }
        print(Indentation.DEC, "","}");
        printWriter.close();
    }

    private void generateSegment(DynaaSegment segment) {
        print("private static TaskSegment create" + StringUtils.capitalize(segment.getName()) + "(final Task task) {");
        print(Indentation.INC, "return new TaskSegment() {");
        print(Indentation.INC, "@Override", "public void run(final TaskContext taskContext) throws Exception {");
        print(Indentation.INC, "");

        for(String s : segment.getBody())
            print(s);

        print(Indentation.DEC, "}");
        print(Indentation.DEC, "};");
        print(Indentation.DEC, "}", "");
    }

    private void generateDataContainer() throws IOException {
        String className = globalVarsContainerClassName;
        printWriter = new PrintWriter(outputDir + className + ".java");
        print("package " + filesPackage + ";", "");
        print(DynaaConsts.HEADER);
        print("public class " + className + " {");
        print(Indentation.INC, "");
        for(Variable var: model.getGlobalVariables().values()) {
            if(var.getExpr() == null) {
                print("private " + var.getType().toString() + " " + var.getName() + ";");
            } else {
                print("private " + var.getType().toString() + " " + var.getName() + " = " + ExprToString.parse(var.getExpr()) + ";");
            }
        }
        print("");
        for(Variable var: model.getGlobalVariables().values()) {
            String cName = StringUtils.capitalize(var.getName());
            print("public get" + cName + "() {");
            print(Indentation.INC, "return " + var.getName() + ";");
            print(Indentation.DEC, "}", "");
            print("public set" + cName + "(" + var.getType().toString() + " " + var.getName() + ") {");
            print(Indentation.INC, "this." + var.getName() + " = " + var.getName() + ";");
            print(Indentation.DEC, "}", "");
        }
        print(Indentation.DEC, "", "}");
        printWriter.close();
    }

    /*******************************************************
     * Utils Methdos
     *******************************************************/

    private void print(Indentation ind, String ... args) {
        if(ind == Indentation.INC)
            ++_indentation;
        if(ind == Indentation.DEC)
            --_indentation;
        print(args);
    }

    private void print(String ... args) {
        for(String str: args) {
            for (int i = 0; i < _indentation; ++i) {
                printWriter.print("    ");
            }
            printWriter.println(str);
        }
    }
}
