package it.sagelab.dynaagen.core;

import it.sagelab.dynaagen.model.Proctype;
import it.sagelab.dynaagen.model.PromelaModel;
import it.sagelab.dynaagen.model.expr.Expression;
import it.sagelab.dynaagen.model.expr.Variable;
import it.sagelab.dynaagen.model.stmnt.*;
import org.apache.commons.lang3.StringUtils;

import java.util.*;

public class DynaaSegmentGenerator implements StatementVisitor {

    private final PromelaModel model;
    private Proctype proctype;
    private Map<Statement, DynaaSegment> segmentMap;
    private int _segCounter = 0;

    public DynaaSegmentGenerator(PromelaModel model) {
        this.model = model;
        segmentMap = new LinkedHashMap<>();
    }

    public Set<DynaaSegment> buildSegments(Proctype proctype) {
        this.proctype = proctype;
        proctype.getSequence().accept(this);
        Set<DynaaSegment> res = new LinkedHashSet<>();
        for(DynaaSegment seg: segmentMap.values())
            res.add(seg);

        return res;
    }

    public DynaaSegment getSegment(Statement statement) {
        return segmentMap.get(statement);
    }

    private DynaaSegment newSegment(String prefix, Statement statement, boolean successEvent) {
        DynaaSegment segment = new DynaaSegment(prefix + "Segment" + _segCounter);
        segmentMap.put(statement, segment);
        if(successEvent)
            segment.addSegment(DynaaSegment.SEGMENT_SUCCESS, null);
        ++_segCounter;
        return segment;
    }

    @Override
    public void visit(StatementSequence sequence) {
        DynaaSegment segment = null;
        if(sequence.getLocalVars().size() != 0) {
            segment = newSegment("init", sequence, true);
            segment.addLine("// initialize task Context");
            for(Variable var :sequence.getLocalVars().values()) {
                segment.addLine("taskContext.put(\"" + var.getName() + "\", " + ExprToString.parse(var.getExpr()) +");");
            }
            segment.addLine("this.scheduleNow(" + DynaaSegment.SEGMENT_SUCCESS + ");");
        }

        for(Statement statement: sequence.getStmnts()) {
            if(statement != null) {
                statement.accept(this);
                DynaaSegment nextSegment = segmentMap.get(statement);

                if (segment != null)
                    segment.appendNext(nextSegment);
                segment = nextSegment;
            }
        }

        if(!sequence.getStmnts().isEmpty())
            segmentMap.putIfAbsent(sequence, segmentMap.get(sequence.getStmnts().get(0)));

    }

    @Override
    public void visit(AssignStmnt assignStmnt) {
        DynaaSegment segment = newSegment("assing", assignStmnt, true);
        Variable var = assignStmnt.getVar();
        if(assignStmnt.getExpression() != null) {
            Set<Variable> variableSet = ExprToString.findVariables(assignStmnt.getExpression());
            variableSet.add(var);
            for(Variable v: variableSet)
                segment.addLine(getVarDeclaration(v));
        } else {
            segment.addLine(getVarDeclaration(var));
        }

        switch (assignStmnt.getOp()) {
            case INCREMENT:
                segment.addLine(setVariableDeclaration(var, var.getName() + " + 1"));
                break;
            case DECREMENT:
                segment.addLine(setVariableDeclaration(var, var.getName() + " - 1"));
                break;
            case ASSIGNMENT:
                segment.addLine(setVariableDeclaration(var, ExprToString.parse(assignStmnt.getExpression())));
                break;
        }
        segment.addLine("this.scheduleNow(" + DynaaSegment.SEGMENT_SUCCESS + ");");
    }

    @Override
    public void visit(IfStmnt ifStmnt) {
        DynaaSegment segment = newSegment("if", ifStmnt, true);
        produceConditionBranching(ifStmnt, segment);
    }

    @Override
    public void visit(DoStmnt doStmnt) {
        DynaaSegment segment = newSegment("do", doStmnt, true);
        produceConditionBranching(doStmnt, segment);

        for(StatementSequence option: doStmnt.getOptions()) {
            segmentMap.get(option).appendNext(segment);
        }
    }

    @Override
    public void visit(ExprStatement exprStatement) {
        DynaaSegment segment = newSegment("expr", exprStatement, true);
        segment.addSegment(DynaaSegment.SEGMENT_BLOCK, segment);
        retrieveVars(segment, exprStatement.getExpression());

        segment.addLine("if( " + ExprToString.parse(exprStatement.getExpression())+ ") {");
        segment.addLine("    this.scheduleNow("+ DynaaSegment.SEGMENT_SUCCESS +");");
        segment.addLine("} else {");
        segment.addLine("    this.scheduleNow(" + DynaaSegment.SEGMENT_BLOCK  + ");");
        segment.addLine("}");
    }

    @Override
    public void visit(KeywordStmnt keywordStmnt) {

    }

    @Override
    public void visit(AssertStmnt assertStmnt) {
        DynaaSegment segment = newSegment("assert", assertStmnt, true);
        retrieveVars(segment, assertStmnt.getExpression());

        segment.addLine("assert(" + ExprToString.parse(assertStmnt.getExpression())+ ");");
        segment.addLine("this.scheduleNow(" + DynaaSegment.SEGMENT_SUCCESS + ");");
    }

    /********************************************
     * Utils
     ********************************************/

    private boolean isElseStmnt(Statement stmnt) {
        return stmnt instanceof KeywordStmnt ? ((KeywordStmnt) stmnt).getKeyword() == KeywordStmnt.Keyword.ELSE : false;
    }

    private void retrieveVars(DynaaSegment segment, Expression expr) {
        Set<Variable> variableSet = ExprToString.findVariables(expr);

        for(Variable var : variableSet) {
            segment.addLine(getVarDeclaration(var));
        }
    }

    private String getVarDeclaration(Variable var) {
        if(model.getGlobalVariables().containsKey(var.getName())) {
            return var.getType() + " " + var.getName() + " = task.get(\"GLOBAL_VARS\").get" + StringUtils.capitalize(var.getName()) + "();";
        } else if(proctype.getParams().containsKey(var.getName())) {
            return var.getType() + " " + var.getName() + " = task.get(\"" + var.getName() + "\");";
        } else {
            return var.getType() + " " + var.getName() + " = taskContext.get(\"" + var.getName() + "\");";
        }
    }

    private String setVariableDeclaration(Variable var, String value) {
        if(model.getGlobalVariables().containsKey(var.getName())) {
            return "task.get(\"GLOBAL_VARS\").set" + StringUtils.capitalize(var.getName()) + "(" + value + ");";
        } else if (proctype.getParams().containsKey(var.getName())) {
            return "task.put(\"" + var.getName() + "\", " + value +");";
        } else {
            return "taskContext.put(\"" + var.getName() + "\", " + value +");";
        }
    }

    private void produceConditionBranching(IfStmnt ifStmnt, DynaaSegment segment) {
        ArrayList<Statement> conditions = new ArrayList<>();
        boolean elseBranch = false;
        for(StatementSequence option: ifStmnt.getOptions()) {
            if(!option.getStmnts().isEmpty()) {
                Statement stmnt = option.getStmnts().get(0);
                conditions.add(stmnt);
                elseBranch |= isElseStmnt(stmnt);
            }
        }
        segment.addLine("ArrayList<Event.EventType> enabledEvents = new ArrayList<>();");
        segment.addLine("Random rand = new Random();");
        segment.addLine("");

        Set<Variable> variableSet = new HashSet<>();
        for(Statement stmnt: conditions) {
            if(stmnt instanceof ExprStatement)
                variableSet.addAll(ExprToString.findVariables(((ExprStatement) stmnt).getExpression()));
        }
        for(Variable var : variableSet) {
            segment.addLine(getVarDeclaration(var));
        }

        int counter = 1;
        for(Statement stmnt: conditions) {
            if(stmnt instanceof ExprStatement) {
                ExprStatement exprStatement = (ExprStatement) stmnt;
                segment.addLine("if(" + ExprToString.parse(exprStatement.getExpression())+ ") {");
                segment.addLine("    enabledEvents.add("+ DynaaSegment.SEGMENT_BRANCH + counter +");");
                segment.addLine("}");
                ++counter;
            } else {
                segment.addLine("enabledEvents.add("+ DynaaSegment.SEGMENT_BRANCH + counter +");");
                ++counter;
            }
        }

        segment.addLine("if(enabledEvents.isEmpty()) {");
        if(elseBranch) {
            segment.addLine("    enabledEvents.add(" + DynaaSegment.SEGMENT_ELSE +");");
        } else {
            segment.addLine("    enabledEvents.add(" + DynaaSegment.SEGMENT_BLOCK +");");
        }
        segment.addLine("}");

        segment.addLine("int _index = rand.nextInt(enabledEvents.size());");
        segment.addLine("this.scheduleNow(enabledEvents.get(_index));");

        counter = 1;
        for(StatementSequence option: ifStmnt.getOptions()) {
            Statement stmnt = option.getStmnts().get(0);
            option.getStmnts().remove(0);
            option.accept(this);
            if(isElseStmnt(stmnt)) {
                segment.addSegment(DynaaSegment.SEGMENT_ELSE, segmentMap.get(option));
            } else {
                segment.addSegment(DynaaSegment.SEGMENT_BRANCH + counter, segmentMap.get(option));
                ++counter;
            }
        }
        if(!elseBranch) {
            segment.addSegment(DynaaSegment.SEGMENT_BLOCK, segment);
        }
    }

}
