package it.sagelab.dynaagen.core;

import it.sagelab.dynaagen.model.Proctype;
import it.sagelab.dynaagen.model.PromelaModel;
import it.sagelab.dynaagen.model.expr.Variable;
import it.sagelab.dynaagen.model.expr.BinaryOpExpr;
import it.sagelab.dynaagen.model.expr.ConstantExpr;
import it.sagelab.dynaagen.model.expr.Expression;
import it.sagelab.dynaagen.model.expr.UnaryOpExpr;
import it.sagelab.dynaagen.model.stmnt.*;
import it.sagelab.dynaagen.parser.PromelaBaseListener;
import it.sagelab.dynaagen.parser.PromelaParser;
import org.antlr.v4.runtime.tree.ParseTreeProperty;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;


public class PromelaReader extends PromelaBaseListener {


    private PromelaModel model;
    private Proctype currentProctype;
    /**  maps nodes to Expressions with Map<ParseTree,Expression>. */
    private final ParseTreeProperty<Expression> exprs = new ParseTreeProperty<>();
    /**  maps nodes to Expressions with Map<ParseTree,Expression>. */
    private final ParseTreeProperty<Statement> statements = new ParseTreeProperty<>();
    private final LinkedList<Map<String, Variable>> varsScope = new LinkedList<>();



    public PromelaReader(){
        model = new PromelaModel();
        varsScope.add(model.getGlobalVariables());
        model.addGlobalVariable(new Variable("pid", Variable.Type.INT, new ConstantExpr("0")));
    }

    public PromelaModel getModel() {
        return model;
    }


    @Override
    public void enterProctype(PromelaParser.ProctypeContext ctx) {
        boolean active = ctx.active() != null;
        currentProctype = new Proctype(ctx.name().getText(), active);
        model.addProctype(currentProctype);
        varsScope.add(currentProctype.getParams());
    }

    @Override
    public void exitProctype(PromelaParser.ProctypeContext ctx) {
        currentProctype.setSequence((StatementSequence) statements.get(ctx.sequence()));
        varsScope.removeLast();
    }

    @Override
    public void exitDeclList(PromelaParser.DeclListContext ctx) {
        Map<String, Variable> decl = varsScope.getLast();
        if(ctx == null || decl == null)
            return;
        for(PromelaParser.OneDeclContext oneDeclContext: ctx.oneDecl()) {
            String typename = oneDeclContext.typename().getText();
            Variable.Type type;
            switch(typename) {
                case "bit":
                case "bool":
                    type = Variable.Type.BOOL;
                    break;
                case "byte":
                    type = Variable.Type.BYTE;
                    break;
                case "short":
                    type = Variable.Type.SHORT;
                    break;
                case "int":
                    type = Variable.Type.INT;
                    break;
                case "mtype":
                    type = Variable.Type.MTYPE;
                    break;
                case "chan":
                    type = Variable.Type.CHAN;
                    break;
                default:
                    throw new UnsupportedOperationException(typename + " type not supported");
            }

            for(PromelaParser.IvarContext ivar: oneDeclContext.ivar()) {
                String varName = ivar.name().getText();
                if(decl.containsKey(varName)) {
                    throw new UnsupportedOperationException("Variable "+ varName + " declared multiple times in the same scope");
                }
                else {
                    Expression expr = null;
                    if(ivar.any_expr() != null)
                        expr = exprs.get(ivar.any_expr());
                    decl.put(varName, new Variable(varName, type, expr));
                }
            }
        }
    }

    /********************************************************************
     * Start Statements Parsing
     ********************************************************************/

    @Override
    public void enterSequence(PromelaParser.SequenceContext ctx) {
        varsScope.add(new HashMap<>());
    }

    @Override
    public void exitSequence(PromelaParser.SequenceContext ctx) {
        Map<String, Variable> localVars = varsScope.removeLast();
        ArrayList<Statement> sequence = new ArrayList<>();
        for(PromelaParser.StepContext stepContext: ctx.step()) {
            if(stepContext.stmnt() != null)
                sequence.add(statements.get(stepContext.stmnt(0)));
        }

        statements.put(ctx, new StatementSequence(sequence, localVars));
    }

    @Override
    public void exitAssignStmnt(PromelaParser.AssignStmntContext ctx) {
        Variable var = (Variable) exprs.get(ctx.assign().varref());
        Expression expr = null;
        AssignStmnt.Op op;
        if(ctx.assign().EQ() != null) {
            expr = exprs.get(ctx.assign().any_expr());
            op = AssignStmnt.Op.ASSIGNMENT;
        } else {
            op = ctx.assign().INC() != null ? AssignStmnt.Op.INCREMENT : AssignStmnt.Op.DECREMENT;
        }
        statements.put(ctx, new AssignStmnt(var, expr, op));
    }

    @Override
    public void exitKeywordStmnt(PromelaParser.KeywordStmntContext ctx) {
        KeywordStmnt.Keyword keyword = null;
        if(ctx.ELSE()  != null) keyword = KeywordStmnt.Keyword.ELSE;
        if(ctx.BREAK() != null) keyword = KeywordStmnt.Keyword.BREAK;
        if(ctx.GOTO()  != null) keyword = KeywordStmnt.Keyword.GOTO;
        KeywordStmnt stmnt = new KeywordStmnt(keyword);
        if(keyword == KeywordStmnt.Keyword.GOTO)
            stmnt.setGotoLabel(ctx.name().getText());
        statements.put(ctx, stmnt);
    }

    @Override
    public void exitLabelStmnt(PromelaParser.LabelStmntContext ctx) {
        Statement stmnt = statements.get(ctx.stmnt());
        stmnt.setLabel(ctx.name().getText());
        statements.put(ctx, stmnt);
    }

    @Override
    public void exitBraketStmnt(PromelaParser.BraketStmntContext ctx) {
        statements.put(ctx, statements.get(ctx.sequence()));
    }

    @Override
    public void exitDoStmnt(PromelaParser.DoStmntContext ctx) {
        ArrayList<StatementSequence> options = new ArrayList<>();
        for(PromelaParser.SequenceContext sequenceContext: ctx.stmnt_options().sequence())
            options.add((StatementSequence) statements.get(sequenceContext));
        statements.put(ctx, new DoStmnt(options));
    }

    @Override
    public void exitIfStmnt(PromelaParser.IfStmntContext ctx) {
        ArrayList<StatementSequence> options = new ArrayList<>();
        for(PromelaParser.SequenceContext sequenceContext: ctx.stmnt_options().sequence())
            options.add((StatementSequence) statements.get(sequenceContext));
        statements.put(ctx, new IfStmnt(options));
    }

    @Override
    public void exitExprStmnt(PromelaParser.ExprStmntContext ctx) {
        statements.put(ctx, new ExprStatement(exprs.get(ctx.expr())));
    }

    @Override
    public void exitAssertStmnt(PromelaParser.AssertStmntContext ctx) {
        statements.put(ctx, new AssertStmnt(exprs.get(ctx.expr())));
    }

    /********************************************************************
     * End Statements Parsing
     ********************************************************************/

    /********************************************************************
     * Start Exprs Parsing
     ********************************************************************/

    @Override
    public void exitBinaryExpr(PromelaParser.BinaryExprContext ctx) {
        Expression leftExpr = exprs.get(ctx.any_expr(0));
        Expression rightExpr = exprs.get(ctx.any_expr(1));
        exprs.put(ctx, new BinaryOpExpr(leftExpr, rightExpr, ctx.binaryOp().getText()));
    }

    @Override
    public void exitUnaryExpr(PromelaParser.UnaryExprContext ctx) {
        Expression expr = exprs.get(ctx.any_expr());
        exprs.put(ctx, new UnaryOpExpr(expr, ctx.unaryOp().getText()));
    }

    @Override
    public void exitBraketExpr(PromelaParser.BraketExprContext ctx) {
        exprs.put(ctx, exprs.get(ctx.any_expr()));
    }

    @Override
    public void exitConstantExpr(PromelaParser.ConstantExprContext ctx) {
        exprs.put(ctx, exprs.get(ctx.constant()));
    }

    @Override
    public void exitVarRefExpr(PromelaParser.VarRefExprContext ctx) {
        exprs.put(ctx, exprs.get(ctx.varref()));
    }

    @Override
    public void exitVarref(PromelaParser.VarrefContext ctx) {
        String varName = "";
        if(ctx.name() != null)
            varName = ctx.name().getText();
        else
            varName = ctx.PID().getText();

        Variable var = null;
        for(int i = varsScope.size() - 1; i >= 0; --i) {
            if(varsScope.get(i).containsKey(varName))
                var = varsScope.get(i).get(varName);
        }
        if(var == null) {
            throw new RuntimeException("Variable " + varName + " not initialized!");
        }
        else {
            exprs.put(ctx, var);
        }
    }

    @Override
    public void exitAnyExpr(PromelaParser.AnyExprContext ctx) {
        // Skip for now
    }

    @Override
    public void exitConstant(PromelaParser.ConstantContext ctx) {
        exprs.put(ctx, new ConstantExpr(ctx.getText()));
    }

    @Override
    public void exitExpr(PromelaParser.ExprContext ctx) {
        if(ctx.any_expr() != null) {
            exprs.put(ctx, exprs.get(ctx.any_expr()));
        } else if(ctx.andorOp() != null) {
            Expression leftExpr = exprs.get(ctx.expr(0));
            Expression rightExpr = exprs.get(ctx.expr(1));
            exprs.put(ctx, new BinaryOpExpr(leftExpr, rightExpr, ctx.andorOp().getText()));
        } else{
            exprs.put(ctx, exprs.get(ctx.expr(0)));
        }
    }

    /********************************************************************
     * End Exprs Parsing
     ********************************************************************/



}
